#!/usr/bin/python2

from Tkinter import *
import os, sys
import math
import bisect
import json
import time


def mktime(s):
    return time.mktime(time.strptime(s, '%Y-%m-%d %H:%M:%S')) / 3600.0

class Cord(object):
  
  def __init__(self, x, y):
    self.x = x
    self.y = y

class Colors(object):

  colors = []

  def __init__(self):
    self.inuse = {}

  def get(self):
    i = 0
    for rgb in self.colors:
      if rgb not in self.inuse:
        self.inuse[rgb] = 1
        return rgb
      i += 1
    # generate new hsv-color
    b = 1 # base of 2^x
    while i >= b: b *= 2
    h = (360.0/b)%360 + 720.0/b*(i-b/2)
    s = 1.0 / 1.1**(b-1); v = s/2 + 0.5
    r,g,b = hsv_to_rgb(h,s,v)
    rgb = '#%02x%02x%02x' % (r*255,g*255,b*255)
    Colors.colors.append(rgb)
    self.inuse[rgb] = 1
    return rgb
   
  def restore(self, color):
    del self.inuse[color]


def hsv_to_rgb(h,s,v):
  # algorithm found at: http://www.cs.rit.edu/~ncs/color/t_convert.html

  if(s == 0) : # achromatic (grey)
    return [v,v,v]
  else:
    h = h/60.0
    i = int(h) # sector 0 to 5
    f = h - i  # factorial part of h
    p = v * ( 1 - s )
    q = v * ( 1 - s * f )
    t = v * ( 1 - s * ( 1 - f ) )
    return [(v,t,p), (q,v,p), (p,v,t), (p,q,v), (t,p,v), (v,p,q)][i]

class TeamList2(Canvas):
  
  def __init__(self, parent, width):
    self = Canvas.__init__(parent)
    self.width = width
    self.vscrollbar = Scrollbar(self, command=self.yview)
    self.vscrollbar.grid(row=0, column=1, sticky=N+S)
    self.hscrollbar = Scrollbar(self, command=self.xview, orient=HORIZONTAL)
    self.hscrollbar.grid(row=1, column=0, sticky=E+W)
    self.config(yscrollcommand=self.vscrollbar.set, 
                xscrollcommand=self.hscrollbar.set)
    self.grid(row=0, column=0, sticky=N+S+E+W)

    # make the canvas expandable
    #parent.grid_rowconfigure(0, weight=1)
    #parent.grid_columnconfigure(0, weight=1)

    #
    # create canvas contents

    self.frame = Frame(self)
    self.frame.rowconfigure(1, weight=1)
    self.frame.columnconfigure(1, weight=1)

    rows = 5 
    for i in range(1,rows):
        for j in range(1,10):
            button = Button(frame, padx=7, pady=7, text="[%d,%d]" % (i,j))
            button.grid(row=i, column=j, sticky='news')

    self.create_window(0, 0, anchor=NW, window=frame)
    self.frame.update_idletasks()
    self.config(scrollregion=canvas.bbox("all"))

class TeamList(Frame):

  def __init__(self, parent, command=None):
    Frame.__init__(self, parent)
    self.scrollbar = Scrollbar(self, orient=VERTICAL)
    self.listbox = Listbox(self, selectmode=EXTENDED, yscrollcommand=self.scrollbar.set)
    self.scrollbar.config(command=self.listbox.yview)
    self.scrollbar.pack(side=RIGHT, fill=Y)
    self.listbox.pack(side=LEFT, fill=BOTH, expand=1)
    self.colors = Colors()
    self.current = ()
    self.poll() # start polling changes
    self.command = command

  def update(self, teams):
    self.listbox.delete(0, END)
    for team in teams:
      self.listbox.insert(END, team.listtext())

  def poll(self):
    now = self.listbox.curselection()
    if now != self.current:
      added = [i for i in now if i not in self.current]
      removed = [i for i in self.current if i not in now]
      self.current = now
      try: now, added, removed = map(int, now), map(int, added), map(int, removed)
      except ValueError: pass
      for i in removed: # colors back
        self.colors.restore(self.color(i))
        self.color(i, '')
      for i in added:
        self.color(i, self.colors.get())
      self.command(now, added=added, removed=removed)
    self.after(250, self.poll)

  def color(self, i, color=None):
    if color:
      self.listbox.itemconfig(i, bg=color)
    else:
      return self.listbox.itemcget(i, 'bg')

START, FINISH = ('start','finish')

class Map(Frame):

  def __init__(self, parent):
    Frame.__init__(self, parent)
    self.grid_rowconfigure(0, weight=1)
    self.grid_columnconfigure(0, weight=1)

    self.xscrollbar = Scrollbar(self, orient=HORIZONTAL)
    self.xscrollbar.grid(row=1, column=0, sticky=E+W)
    self.yscrollbar = Scrollbar(self, orient=VERTICAL)
    self.yscrollbar.grid(row=0, column=1, sticky=N+S)

    self.canvas = Canvas(self, xscrollcommand=self.xscrollbar.set,
                               yscrollcommand=self.yscrollbar.set)
    self.canvas.grid(row=0, column=0, sticky=N+S+E+W)
    self.xscrollbar.config(command=self.canvas.xview)
    self.yscrollbar.config(command=self.canvas.yview)
    
class Time(Scale):

  limits = [3.0, 8.0, 16.0, 24.0]

  def __init__(self, parent, (timemin, timemax), limit=None, change=None):
    timemin = math.floor(timemin)
    timemax = math.ceil(timemax)
    Scale.__init__(self, parent, orient=HORIZONTAL, showvalue=1, from_=timemin, to=timemax, resolution=0.05,
                   command=self.handler)
    self.change = change # callback when changed 
    self.set(limit)
    self.current = self.get()

  def handler(self, event):
    if self.current != self.get():
      self.current = self.get()
      if self.change:
        self.change(self.current)

class ControlPoint(object):

  def __repr__(self):
    return 'P '+self.name

  def __init__(self, id, cord, name=None):
    self.id = id
    self.cord = cord
    self.name = name or id
    self.flag = None

  COLOR="#f0f"
  RAD=10
  FINISH_RAD=6

  def draw(self, canvas, scale):
    x,y = self.cord.x * scale / 100, self.cord.y * scale / 100
    x0, y0, x1, y1 = x-self.RAD, y-self.RAD, x+self.RAD, y+self.RAD
    try:
      canvas.coords(self._line, x0, y0, x1, y1)
    except AttributeError:
      self._line = canvas.create_oval(x0, y0, x1, y1,
                                      outline=self.COLOR, width=2)
    #if self.name.isdigit():
    if not self.flag:
      x0, y0 = x+self.RAD, y-self.RAD/2
      try:
        canvas.coords(self._text, x0, y0)
      except AttributeError:
        self._text = canvas.create_text(x0, y0,
                                        #anchor=SW, text=str(self.name), fill=self.COLOR)
                                        anchor=SW, text=self.name, fill=self.COLOR)
    if self.flag == FINISH:
      x0, y0, x1, y1 = x-self.FINISH_RAD, y-self.FINISH_RAD, x+self.FINISH_RAD, y+self.FINISH_RAD
      try:
        canvas.coords(self._line2, x0, y0, x1, y1)
      except AttributeError:
        self._line2 = canvas.create_oval(x0, y0, x1, y1,
                                         outline=self.COLOR, width=2)

class Team(object):

  def __init__(self, nbr, name, route=[], id=0):
    self.id = id
    self.nbr = nbr
    self.name = name
    self.route = route # [('start', time), ('point', time), ...]
    self.lengths = None
    self.color = None

  def listtext(self):
    return self.nbr + '. ' + self.name

  def __cmp__(self, other):
    return cmp(self.id, other.id)

  def bind(self, legmap, color=None):
    self.color = color
    self.lengths = {}

  def unbind(self, legmap):
    for leg in self.lengths.iterkeys():
      legmap.get(leg).unbind(self)
    self.color = None
    self.lengths = None

  def settime(self, time, legmap):
    lengths = {}
    p0, t0 = self.route[0]
    for p,t in self.route[1:]:
      if time <= t0: break
      leg = (p0,p)
      length = (time >= t) and 1.0 or ((time-t0) / (t-t0))
      try:
        assert length == self.lengths[leg]
        del self.lengths[leg]
      except KeyError: # didn't exist
        legmap.get(leg).bind(self)
      except AssertionError: # changed
        legmap.get(leg).update()
        del self.lengths[leg]
      lengths[leg] = length
      p0, t0 = p, t
    for leg in self.lengths.keys():
      legmap.get(leg).unbind(self)
    self.lengths = lengths

ascend = lambda a,b: (a<=b) and (a,b) or (b,a)

class Leg(list):

  WIDTH = 3

  def __init__(self, id, start, end, _update):
    self.id = id
    self.start = start
    self.end = end
    self.length = math.sqrt((end.x-start.x)**2.0 + (end.y-start.y)**2.0)
    if self.length:
        self.vector = Cord((end.x-start.x)/self.length, (end.y-start.y)/self.length)
    else:
        self.vector = Cord(start.x, end.x)
    self._update = _update # update function
    self._lines = {}
  
  def bind(self, team):
    bisect.insort(self, team)
    self._update(self.id)

  def unbind(self, team):
    self.remove(team)
    self._update(self.id)

  def update(self):
    self._update(self.id)

  def draw(self, canvas, width, scale):
    if not self.length:
      return
    _lines = {}
    pos = (width-1)/2.0 - (len(self)-1)
    X0 = self.start.x*scale/100 + self.vector.x*ControlPoint.RAD
    Y0 = self.start.y*scale/100 + self.vector.y*ControlPoint.RAD
    X1 = self.end.x*scale/100
    Y1 = self.end.y*scale/100
    hidden_start = ControlPoint.RAD / self.length*scale/100 # 0 .. 1.0
    for team in self:
      length = team.lengths[self.id]
      if length > hidden_start:
        x0 = X0 - self.vector.y*Leg.WIDTH*pos
        y0 = Y0 + self.vector.x*Leg.WIDTH*pos
        if length == 1.0:
          x1 = X1 - self.vector.y*Leg.WIDTH*pos
          y1 = Y1 + self.vector.x*Leg.WIDTH*pos
        else:
          x1 = X0*(1.0-length) + X1*length - self.vector.y*Leg.WIDTH*pos
          y1 = Y0*(1.0-length) + Y1*length + self.vector.x*Leg.WIDTH*pos
        try: 
          line = self._lines[team]
          canvas.coords(line, x0,y0,x1,y1)
        except KeyError:
          line = canvas.create_line(x0,y0,x1,y1, fill=team.color, width=Leg.WIDTH)
        else:
          del self._lines[team]
        _lines[team] = line
      pos += 1
    for line in self._lines.values():
      canvas.delete(line)    
    self._lines = _lines

class Legmap(dict):
  
  def __init__(self, controlpoints):
    self.controlpoints = controlpoints
    self._update = {}

  def get(self, leg):
    a,b = leg
    p0 = self.controlpoints[a]
    p1 = self.controlpoints[b]
    return self.setdefault(leg, Leg(leg, p0.cord, p1.cord, self.update))
  
  def draw(self, canvas):
    while self._update:
      (a,b),value = self._update.popitem()
      leg, leg2 = self.get((a,b)), self.get((b,a))
      try:
        del self._update[(b,a)]
      except KeyError:
        pass
      width = len(leg) + len(leg2)
      leg.draw(canvas, width, self.scale)
      leg2.draw(canvas, width, self.scale)

  def redraw(self, scale, canvas):
    self.scale = scale
    for leg in self.keys():
      self._update[leg] = 1
    self.draw(canvas)

  def update(self, leg):
    self._update[leg] = 1

minx = lambda *tup: min([x for x in tup if x != None])
  
class Prog(Frame):
   
  def __init__(self, parent, radafile):
    Frame.__init__(self, parent)

    self.readfiles(radafile)

    self.grid_rowconfigure(2, weight=1)
    self.grid_columnconfigure(1, weight=1)

    self.scale = IntVar()
    self.scale.set(100)

    self.menubar = Menu(self)

    menu1 = Menu(self.menubar, tearoff=0)
    menu1.add_command(label='Exit', command=parent.quit)
    self.menubar.add_cascade(label='File', menu=menu1)
 
    menu2 = Menu(self.menubar, tearoff=0)
    for s in (100,50,25):
      menu2.add_radiobutton(label=str(s)+'%', variable=self.scale, value=s, command=self.update_scale)
    self.menubar.add_cascade(label='Map', menu=menu2)

    parent.config(menu=self.menubar)

    self.teamlist = TeamList(self, command=self.update_selected)
    self.teamlist.update(self.teams)
    self.teamlist.grid(row=1, column=0, rowspan=2, sticky=N+S)

    if not self.timelimit:
      try:
        self.timelimit = Time.limits[bisect.bisect_right(Time.limits, self.timerange[1])-1]
      except IndexError:
        self.timelimit = 0.0

    self.time = Time(self, self.timerange, limit=self.timelimit, change=self.update_time)
    self.time.grid(row=1, column=1, sticky=W+E)
    
    self.map = Map(self)
    self.map.grid(row=2, column=1, sticky=N+W+S+E)
    self.canvas = self.map.canvas

    self.legmap = Legmap(self.controlpoints)

    self.update_scale()
    self.update_time(self.timelimit)
    
  def readfiles(self, radafile):
    if os.path.splitext(radafile)[1] == '.rada':
        self.read_taok(radafile)
    elif os.path.splitext(radafile)[1] == '.json':
        self.read_gpspunch(radafile)

  def _init(self):
    self.teams = []
    self.selected = []
    self.controlpoints = {}
    self.timerange = (0.0, 0.0)
    self.timelimit = None

  def read_gpspunch(self, jsonfile):
    """http://gps-punch.appspot.com/qr/qr/rastitupa.json?e=<kilp-id>
       + {"kartta": {"file":"filepath.png",
                     "trasform_x": [lat, lon, off],
                     "trasform_y": [lat, lon, off]}}
    """
    self._init()
    data = json.load(open(jsonfile))
    kartta = data['kartta']

    self.mapfile = kartta['file']
    self.original_image = PhotoImage(file = self.mapfile)

    tr_x, tr_y = kartta['transform_x'], kartta['transform_y']
    to_map_y = lambda lat, lon: lat * tr_y[0] + lon * tr_y[1] + tr_y[2]
    to_map_x = lambda lat, lon: lat * tr_x[0] + lon * tr_x[1] + tr_x[2]

    self.timelimit = 8.0 # FIXME

    # Rastit
    for rasti in data['rastit']:
      lat, lon = float(rasti['lat']), float(rasti['lon'])
      x, y = to_map_x(lat, lon), to_map_y(lat, lon)
      controlpoint = ControlPoint(str(rasti['id']),Cord(x,y),name=rasti['koodi'])
      self.controlpoints[controlpoint.id] = controlpoint
      if controlpoint.name == 'MAALI':
        self.controlpoints[FINISH] = controlpoint

    leimat_joukkueittain = {}
    for leima in data['tupa']:
      leimat_joukkueittain.setdefault(leima['joukkue'], []).append(leima)

    start = self.controlpoints.get(START) or self.controlpoints[FINISH]
    finish = self.controlpoints[FINISH]

    sarjat = {s['id']: s for s in data['sarjat']}

    # Joukkueet
    for joukkue in data['joukkueet']:
      nbr, nimi = joukkue['nimi'].split(None, 1)

      alkuaika = mktime(sarjat[joukkue['sarja']]['alkuaika'])

      route = [(start.id, 0.0)]

      leimat = leimat_joukkueittain.get(joukkue['id']) or []
      leimattu = set()

      for leima in leimat:
        koodi = unicode(leima['rasti'])
        if koodi in self.controlpoints and koodi not in leimattu:
          aika = mktime(leima['aika'])
          route.append((koodi, aika - alkuaika))
          leimattu.add(koodi)

      if joukkue['last']:
          loppuaika = mktime(joukkue['last'])
          route.append((finish.id, loppuaika-alkuaika))

      self.teams.append(Team(nbr, nimi, route, id=joukkue['id']))
      
      mi, ma = min(aika for koodi, aika in route), max(aika for koodi, aika in route)
      if self.timelimit:
         ma = min(ma, self.timelimit*1.5)
      self.timerange = (minx(self.timerange[0], mi), max(self.timerange[1], ma))

  def read_taok(self, radafile):
    self._init()

    radafile = os.path.expanduser(radafile)
    resfile = os.path.splitext(radafile)[0] + '.aeg'
 
    for line in open(radafile).readlines():
      line = line.rstrip('\n\r\t')
      if line.startswith('map='):
        self.mapfile = line.split('=',1)[1]
      elif line.startswith('dpi='):
        self.dpi = float(line.split('=',1)[1])
      elif line.startswith('mastap='):
        self.mapscale = float(line.split('=',1)[1])
      elif line.startswith('limit='):
        self.timelimit = float(line.split('=',1)[1])
      elif line.split('=',1)[0].isdigit():
        x,y,flags,name = line.split('=',1)[1].split('|',3)
        x,y = map(int, (x,y))
        name = name.replace('KP', '')
        controlpoint = ControlPoint(name,Cord(x,y))
        self.controlpoints[controlpoint.name] = controlpoint

        if flags == '0':
          controlpoint.flag = START
          self.controlpoints[START] = controlpoint
        if flags == '6':
          controlpoint.flag = FINISH
          self.controlpoints[FINISH] = controlpoint

    self.original_image = PhotoImage(file = self.mapfile)

    for line in open(resfile).readlines():
      self.teams.append(self._parseteam(line, count=len(self.teams)))

  def _parseteam(self, line, count):
    parts = line.rstrip('\n\r\t').split('\t')
    nbr, name = parts[0:2]
    try:
      route = map(self._parsepoint, parts[2:])
    except ValueError: # Old format
      route = [(self.controlpoints[p].name,0.0) for p in parts[2:]]
      start = self.controlpoints.get(START) or self.controlpoints[FINISH]
      finish = self.controlpoints[FINISH]
      route.insert(0, (start.name, 0.0))
      route.append((finish.name, 0.0))
    return Team(nbr, name, route, id=count)

  def _parsepoint(self, txt):
    p,t = txt.split('|')
    h,m,s = map(int, t.split(':'))
    hours = h + m/60.0 + s /3600.0
    self.timerange = (minx(self.timerange[0], hours), max(self.timerange[1], hours))
    return self.controlpoints[p].name, hours

  def update_scale(self):
    scale = self.scale.get()
 
    self._image = self.original_image.subsample(100/scale)
    try:
      self.canvas.itemconfig(self._imageid, image=self._image)
    except AttributeError:
      self._imageid = self.canvas.create_image(0,0, image=self._image, anchor=NW)
    self.canvas.config(scrollregion=(0,0, self._image.width(), self._image.height()))

    for controlpoint in self.controlpoints.values():
      controlpoint.draw(self.canvas, scale)

    self.legmap.redraw(scale, self.canvas)

  def update_time(self, time):
    self.time = time
    for team in self.selected:
      team.settime(time, self.legmap)
    self.legmap.draw(self.canvas)

  def update_selected(self, list, added=[], removed=[]):
    for i in removed:
      self.teams[i].unbind(self.legmap)
    for i in added:
      team = self.teams[i]
      color = self.teamlist.color(i)
      team.bind(self.legmap, color=color)
      team.settime(self.time, self.legmap)
    self.selected = [self.teams[i] for i in list]
    self.legmap.draw(self.canvas)
  
assert len(sys.argv) == 2, 'usage: rogain rada-file'
root = Tk()
root.geometry("%dx%d" % (root.winfo_screenwidth(), root.winfo_screenheight()))

prog = Prog(root, sys.argv[1])
prog.pack(side=LEFT, fill=BOTH, expand=YES)
root.mainloop()
